'''
# triangulate.py
#
# Created by Danny Wynne
'''
import maya.OpenMaya as om
import maya.cmds as cmds
import math
class Triangulate(object):
	def __init__(self,origin, ref_pnts, params, base_vecs, debug=False):
		''' 
		(MPoint)origin - where the target point is at its rest point
		(TriData)ref_pnts - 3 MPoints that define the new space of the target
		(TriData)params - 3 floats that are parameter t in a line equation Line = pnt + t(Vector)
		(TriData)base_vecs - 3 MVectors that define the original space of the target
		
		there are three components to this triangulation algorithm: a,b,c
		A - is a plane equation
		B - is a ray that is cast towards the plane
		C - a ray
		
		the B ray is swept across the A plane until the 
			dot of C ray's vector and the vector formed by (B A intersection) - (C ray's base point)
			flips from positive to negative or vice versa
			
		the algorithm sweeps the B ray 360 degrees with large steps until the it has a hit,
			then the algorithm will run again for just the range of degrees where it found a hit
			if no hit is found, it will switch the A and C components and try again. This could happen
			if the A plane and B ray are too close to parallel. the ray/plane intersection avoids calculating if 
			the A plane and B ray are too close to parallel to avoid points at extreme distances
			
		'''
		self.debug = debug
		a_pnt = origin+((ref_pnts.a - origin) * params.a)
		a_vec = base_vecs.a
		a_vec.normalize()
		
		b_pnt = origin+((ref_pnts.b - origin)*params.b)
		b_normal = base_vecs.b^base_vecs.a
		b_normal.normalize()
		
		c_pnt = origin+((ref_pnts.c-origin)*params.c)
		c_vec = base_vecs.c
		c_vec.normalize()
		
		self.a_point_on_plane = a_pnt
		self.a_normal = a_vec
		
		self.b_pnt = b_pnt
		self.b_intersect_vector = b_normal
		self.b_axis = base_vecs.b
		
		self.c_pnt = c_pnt
		self.c_vec = c_vec
		
		self.iter = 50
		self.inc = 6.28/self.iter
		self.n_offset = 0
		
		self.n, self.prev_n = None, None
		self.pnt = None
		self.offset = 0
		self.algorithm(self.debug)
		if self.n == None:
			#switch the A and C components
			self.a_point_on_plane = c_pnt
			self.a_normal = c_vec
			self.c_pnt = a_pnt
			self.c_vec = a_vec
			self.algorithm()
		if self.n == None: return None
		
		#run the algorithm again with a tighter range
		self.n_offset = self.prev_n
		self.inc = (self.n-self.prev_n)/self.iter
		self.algorithm()
		
		if self.debug and cmds.objExists("debug_grp"):
			res = cmds.polyPlane(ax=(self.a_normal.x, self.a_normal.y, self.a_normal.z), w=10, h=10, sx=1, sy=1, ch=0, name="a_plane")[0]
			cmds.setAttr(res+".t", self.a_point_on_plane.x, self.a_point_on_plane.y, self.a_point_on_plane.z)
			cmds.parent(res, "debug_grp")
			res = cmds.polyPlane(ax=(self.c_vec.x, self.c_vec.y, self.c_vec.z), w=10, h=10, sx=1, sy=1, ch=0, name="c_plane")[0]
			cmds.setAttr(res+".t", self.c_pnt.x, self.c_pnt.y, self.c_pnt.z)
			cmds.parent(res, "debug_grp")

	def algorithm(self, debug=0):
		iter = self.iter
		inc = self.inc
		quat = om.MQuaternion()
		toggle = None
		n = 0
		for i in range(iter):
			prev_n = n
			n = (i*inc)+self.n_offset
			quat.setAxisAngle(self.b_axis, n)
			transformed_intersect_vector = self.b_intersect_vector*quat.asMatrix()
			line_pnt_a, line_pnt_b = self.b_pnt, self.b_pnt+transformed_intersect_vector
			
			intersection_result = line_plane_intersection(line_pnt_a, line_pnt_b, self.a_point_on_plane, self.a_normal, .2)
			if intersection_result == None: continue
			t = intersection_result[1]
			if t<0: continue
			intersection_pnt = intersection_result[0]
			
			#debug
			if cmds.objExists("debug_grp") and debug:
				crv = cmds.curve(p=((line_pnt_a.x, line_pnt_a.y, line_pnt_a.z), (intersection_pnt.x, intersection_pnt.y, intersection_pnt.z)), d=1)
				cmds.setAttr(crv+".overrideEnabled", 1)
				cmds.setAttr(crv+".overrideColor", 3)
				cmds.parent(crv, "debug_grp")
				
			if type(intersection_pnt).__name__ != "MPoint": continue
			test_vec = (intersection_pnt-self.c_pnt)
			test_vec.normalize()
			dot = test_vec*self.c_vec
			if toggle == None: toggle = dot
			if toggle < 0.0:
				if dot > 0.0:
					self.n, self.prev_n = n, prev_n
					self.pnt = intersection_pnt
					return
			if toggle > 0.0:
				if dot<0.0:
					self.n, self.prev_n = n, prev_n
					self.pnt = intersection_pnt
					return
class TriData():
	def __init__(self, a,b,c):
		self.a, self.b, self.c = a,b,c

def line_plane_intersection(A,B, point_on_plane, normal, tol=.01):
	dot = (B-A)*normal
	if dot < tol and dot > -tol: return None
	t = normal*(point_on_plane-A)/(normal*(B-A))
	P = A+(B-A)*t
	return P, t
	
def get_ref_params(origin, target, ref_pnts):
	param_a = get_ref_param(origin, target, ref_pnts.a)
	param_b = get_ref_param(origin, target, ref_pnts.b)
	param_c = get_ref_param(origin, target, ref_pnts.c)
	params = TriData(param_a, param_b, param_c)
	return params

def get_ref_param(pnt_a, pnt_b, pnt_c):
	'''
	a line segment can be expressed parametrically as as Line = pnt + t(Vector)
	this function adjusts pnt_c so that it creates a right angle, 
	and returns parameter t from the equation pnt_a+t*(pnt_c-pnt_a)
	'''
	vec_hyp, vec_adj = (pnt_b-pnt_a), (pnt_c-pnt_a)
	angle_ab = vec_hyp.angle(vec_adj)
	hyp_length = pnt_a.distanceTo(pnt_b)
	adj_length = math.cos(angle_ab)*hyp_length
	vec_adj.normalize()
	adjusted_pnt_c = pnt_a+(vec_adj*adj_length)
	#test wether the adjusted_pnt_c is on vec_adj ray, or if it went the other direction
	testa, testb = pnt_c-pnt_a, adjusted_pnt_c-pnt_a
	testa.normalize();testb.normalize()
	neg = testa*testb
	total_dist = pnt_a.distanceTo(pnt_c)
	ref_dist = pnt_a.distanceTo(adjusted_pnt_c)
	t = ref_dist/total_dist
	return t*neg
	
					